@@ -4,21 +4,34 @@ import (
 	"encoding/json"
 	"fmt"
 	"strings"
+	"sync"
 
+	"github.com/sirupsen/logrus"
 	"golang.org/x/net/context"
 	api "k8s.io/api/core/v1"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
+	"k8s.io/apimachinery/pkg/util/intstr"
 	"k8s.io/client-go/kubernetes"
-	// Register all available authentication methods
-	_ "k8s.io/client-go/plugin/pkg/client/auth"
+	_ "k8s.io/client-go/plugin/pkg/client/auth" // Register all available authentication methods
 
 	"gitlab.com/gitlab-org/gitlab-runner/common"
 	"gitlab.com/gitlab-org/gitlab-runner/executors"
+	"gitlab.com/gitlab-org/gitlab-runner/helpers/container/helperimage"
+	"gitlab.com/gitlab-org/gitlab-runner/helpers/dns"
+	"gitlab.com/gitlab-org/gitlab-runner/session/proxy"
+)
+
+const (
+	buildContainerName  = "build"
+	helperContainerName = "helper"
 )
 
 var (
 	executorOptions = executors.ExecutorOptions{
-		SharedBuildsDir: false,
+		DefaultCustomBuildsDirEnabled: true,
+		DefaultBuildsDir:              "/builds",
+		DefaultCacheDir:               "/cache",
+		SharedBuildsDir:               false,
 		Shell: common.ShellScriptInfo{
 			Shell:         "bash",
 			Type:          common.NormalShell,
@@ -40,6 +53,7 @@ type executor struct {
 	pod         *api.Pod
 	credentials *api.Secret
 	options     *kubernetesOptions
+	services    []api.Service
 
 	configurationOverwrites *overwrites
 	buildLimits             api.ResourceList
@@ -49,6 +63,18 @@ type executor struct {
 	serviceRequests         api.ResourceList
 	helperRequests          api.ResourceList
 	pullPolicy              common.KubernetesPullPolicy
+
+	helperImageInfo helperimage.Info
+}
+
+type serviceDeleteResponse struct {
+	serviceName string
+	err         error
+}
+
+type serviceCreateResponse struct {
+	service *api.Service
+	err     error
 }
 
 func (s *executor) setupResources() error {
@@ -56,28 +82,28 @@ func (s *executor) setupResources() error {
 
 	// Limit
 	if s.buildLimits, err = limits(s.Config.Kubernetes.CPULimit, s.Config.Kubernetes.MemoryLimit); err != nil {
-		return fmt.Errorf("invalid build limits specified: %s", err.Error())
+		return fmt.Errorf("invalid build limits specified: %w", err)
 	}
 
 	if s.serviceLimits, err = limits(s.Config.Kubernetes.ServiceCPULimit, s.Config.Kubernetes.ServiceMemoryLimit); err != nil {
-		return fmt.Errorf("invalid service limits specified: %s", err.Error())
+		return fmt.Errorf("invalid service limits specified: %w", err)
 	}
 
 	if s.helperLimits, err = limits(s.Config.Kubernetes.HelperCPULimit, s.Config.Kubernetes.HelperMemoryLimit); err != nil {
-		return fmt.Errorf("invalid helper limits specified: %s", err.Error())
+		return fmt.Errorf("invalid helper limits specified: %w", err)
 	}
 
 	// Requests
 	if s.buildRequests, err = limits(s.Config.Kubernetes.CPURequest, s.Config.Kubernetes.MemoryRequest); err != nil {
-		return fmt.Errorf("invalid build requests specified: %s", err.Error())
+		return fmt.Errorf("invalid build requests specified: %w", err)
 	}
 
 	if s.serviceRequests, err = limits(s.Config.Kubernetes.ServiceCPURequest, s.Config.Kubernetes.ServiceMemoryRequest); err != nil {
-		return fmt.Errorf("invalid service requests specified: %s", err.Error())
+		return fmt.Errorf("invalid service requests specified: %w", err)
 	}
 
 	if s.helperRequests, err = limits(s.Config.Kubernetes.HelperCPURequest, s.Config.Kubernetes.HelperMemoryRequest); err != nil {
-		return fmt.Errorf("invalid helper requests specified: %s", err.Error())
+		return fmt.Errorf("invalid helper requests specified: %w", err)
 	}
 	return nil
 }
@@ -110,7 +136,7 @@ func (s *executor) Prepare(options common.ExecutorPrepareOptions) (err error) {
 	}
 
 	if s.kubeClient, err = getKubeClient(options.Config.Kubernetes, s.configurationOverwrites); err != nil {
-		return fmt.Errorf("error connecting to Kubernetes: %s", err.Error())
+		return fmt.Errorf("error connecting to Kubernetes: %w", err)
 	}
 
 	s.Println("Using Kubernetes executor with image", s.options.Image.Name, "...")
@@ -133,18 +159,26 @@ func (s *executor) Run(cmd common.ExecutorCommand) error {
 		}
 	}
 
-	containerName := "build"
+	containerName := buildContainerName
 	containerCommand := s.BuildShell.DockerCommand
 	if cmd.Predefined {
-		containerName = "helper"
-		containerCommand = common.ContainerCommandBuild
+		containerName = helperContainerName
+		containerCommand = s.helperImageInfo.Cmd
 	}
 
 	ctx, cancel := context.WithCancel(context.Background())
 	defer cancel()
 
+	s.Debugln(fmt.Sprintf(
+		"Starting in container %q the command %q with script: %s",
+		containerName,
+		containerCommand,
+		cmd.Script,
+	))
+
 	select {
 	case err := <-s.runInContainer(ctx, containerName, containerCommand, cmd.Script):
+		s.Debugln(fmt.Sprintf("Container %q exited with error: %v", containerName, err))
 		if err != nil && strings.Contains(err.Error(), "command terminated with exit code") {
 			return &common.BuildError{Inner: err}
 		}
@@ -168,25 +202,65 @@ func (s *executor) Cleanup() {
 			s.Errorln(fmt.Sprintf("Error cleaning up secrets: %s", err.Error()))
 		}
 	}
+
+	ch := make(chan serviceDeleteResponse)
+	var wg sync.WaitGroup
+	wg.Add(len(s.services))
+
+	for _, service := range s.services {
+		go s.deleteKubernetesService(service.ObjectMeta.Name, ch, &wg)
+	}
+
+	go func() {
+		wg.Wait()
+		close(ch)
+	}()
+
+	for res := range ch {
+		if res.err != nil {
+			s.Errorln(fmt.Sprintf("Error cleaning up the pod service %q: %v", res.serviceName, res.err))
+		}
+	}
+
 	closeKubeClient(s.kubeClient)
 	s.AbstractExecutor.Cleanup()
 }
 
-func (s *executor) buildContainer(name, image string, imageDefinition common.Image, requests, limits api.ResourceList, command ...string) api.Container {
+func (s *executor) deleteKubernetesService(serviceName string, ch chan<- serviceDeleteResponse, wg *sync.WaitGroup) {
+	defer wg.Done()
+
+	err := s.kubeClient.CoreV1().Services(s.configurationOverwrites.namespace).Delete(serviceName, &metav1.DeleteOptions{})
+	ch <- serviceDeleteResponse{serviceName: serviceName, err: err}
+}
+
+func (s *executor) buildContainer(name, image string, imageDefinition common.Image, requests, limits api.ResourceList, containerCommand ...string) api.Container {
 	privileged := false
+	containerPorts := make([]api.ContainerPort, len(imageDefinition.Ports))
+	proxyPorts := make([]proxy.Port, len(imageDefinition.Ports))
+
+	for i, port := range imageDefinition.Ports {
+		proxyPorts[i] = proxy.Port{Name: port.Name, Number: port.Number, Protocol: port.Protocol}
+		containerPorts[i] = api.ContainerPort{ContainerPort: int32(port.Number)}
+	}
+
+	if len(proxyPorts) > 0 {
+		serviceName := imageDefinition.Alias
+
+		if serviceName == "" {
+			serviceName = name
+			if name != buildContainerName {
+				serviceName = fmt.Sprintf("proxy-%s", name)
+			}
+		}
+
+		s.ProxyPool[serviceName] = s.newProxy(serviceName, proxyPorts)
+	}
+
 	if s.Config.Kubernetes != nil {
 		privileged = s.Config.Kubernetes.Privileged
 	}
 
-	if len(command) == 0 && len(imageDefinition.Command) > 0 {
-		command = imageDefinition.Command
-	}
-
-	var args []string
-	if len(imageDefinition.Entrypoint) > 0 {
-		args = command
-		command = imageDefinition.Entrypoint
-	}
+	command, args := s.getCommandAndArgs(imageDefinition, containerCommand...)
 
 	return api.Container{
 		Name:            name,
@@ -199,6 +273,7 @@ func (s *executor) buildContainer(name, image string, imageDefinition common.Ima
 			Limits:   limits,
 			Requests: requests,
 		},
+		Ports:        containerPorts,
 		VolumeMounts: s.getVolumeMounts(),
 		SecurityContext: &api.SecurityContext{
 			Privileged: &privileged,
@@ -207,13 +282,23 @@ func (s *executor) buildContainer(name, image string, imageDefinition common.Ima
 	}
 }
 
-func (s *executor) getVolumeMounts() (mounts []api.VolumeMount) {
-	path := strings.Split(s.Build.BuildDir, "/")
-	path = path[:len(path)-1]
+func (s *executor) getCommandAndArgs(imageDefinition common.Image, command ...string) ([]string, []string) {
+	if len(command) == 0 && len(imageDefinition.Entrypoint) > 0 {
+		command = imageDefinition.Entrypoint
+	}
 
+	var args []string
+	if len(imageDefinition.Command) > 0 {
+		args = imageDefinition.Command
+	}
+
+	return command, args
+}
+
+func (s *executor) getVolumeMounts() (mounts []api.VolumeMount) {
 	mounts = append(mounts, api.VolumeMount{
 		Name:      "repo",
-		MountPath: strings.Join(path, "/"),
+		MountPath: s.Build.RootDir,
 	})
 
 	for _, mount := range s.Config.Kubernetes.Volumes.HostPaths {
@@ -349,6 +434,10 @@ type dockerConfigEntry struct {
 	Username, Password string
 }
 
+func (s *executor) projectUniqueName() string {
+	return dns.MakeRFC1123Compatible(s.Build.ProjectUniqueName())
+}
+
 func (s *executor) setupCredentials() error {
 	authConfigs := make(map[string]dockerConfigEntry)
 
@@ -373,7 +462,7 @@ func (s *executor) setupCredentials() error {
 	}
 
 	secret := api.Secret{}
-	secret.GenerateName = s.Build.ProjectUniqueName()
+	secret.GenerateName = s.projectUniqueName()
 	secret.Namespace = s.configurationOverwrites.namespace
 	secret.Type = api.SecretTypeDockercfg
 	secret.Data = map[string][]byte{}
@@ -393,7 +482,9 @@ func (s *executor) setupBuildPod() error {
 		services[i] = s.buildContainer(fmt.Sprintf("svc-%d", i), resolvedImage, service, s.serviceRequests, s.serviceLimits)
 	}
 
-	labels := make(map[string]string)
+	// We set a default label to the pod. This label will be used later
+	// by the services, to link each service to the pod
+	labels := map[string]string{"pod": s.projectUniqueName()}
 	for k, v := range s.Build.Runner.Kubernetes.PodLabels {
 		labels[k] = s.Build.Variables.ExpandValue(v)
 	}
@@ -412,10 +503,28 @@ func (s *executor) setupBuildPod() error {
 		imagePullSecrets = append(imagePullSecrets, api.LocalObjectReference{Name: s.credentials.Name})
 	}
 
+	podConfig := s.preparePodConfig(labels, annotations, services, imagePullSecrets)
+
+	pod, err := s.kubeClient.CoreV1().Pods(s.configurationOverwrites.namespace).Create(&podConfig)
+	if err != nil {
+		return err
+	}
+
+	s.pod = pod
+	s.services, err = s.makePodProxyServices()
+	if err != nil {
+		return err
+	}
+
+	return nil
+}
+
+func (s *executor) preparePodConfig(labels, annotations map[string]string, services []api.Container, imagePullSecrets []api.LocalObjectReference) api.Pod {
 	buildImage := s.Build.GetAllVariables().ExpandValue(s.options.Image.Name)
-	pod, err := s.kubeClient.CoreV1().Pods(s.configurationOverwrites.namespace).Create(&api.Pod{
+
+	return api.Pod{
 		ObjectMeta: metav1.ObjectMeta{
-			GenerateName: s.Build.ProjectUniqueName(),
+			GenerateName: s.projectUniqueName(),
 			Namespace:    s.configurationOverwrites.namespace,
 			Labels:       labels,
 			Annotations:  annotations,
@@ -425,23 +534,89 @@ func (s *executor) setupBuildPod() error {
 			ServiceAccountName: s.configurationOverwrites.serviceAccount,
 			RestartPolicy:      api.RestartPolicyNever,
 			NodeSelector:       s.Config.Kubernetes.NodeSelector,
+			Tolerations:        s.Config.Kubernetes.GetNodeTolerations(),
 			Containers: append([]api.Container{
 				// TODO use the build and helper template here
-				s.buildContainer("build", buildImage, s.options.Image, s.buildRequests, s.buildLimits, s.BuildShell.DockerCommand...),
-				s.buildContainer("helper", s.Config.Kubernetes.GetHelperImage(), common.Image{}, s.helperRequests, s.helperLimits, s.BuildShell.DockerCommand...),
+				s.buildContainer(buildContainerName, buildImage, s.options.Image, s.buildRequests, s.buildLimits, s.BuildShell.DockerCommand...),
+				s.buildContainer(helperContainerName, s.getHelperImage(), common.Image{}, s.helperRequests, s.helperLimits, s.BuildShell.DockerCommand...),
 			}, services...),
 			TerminationGracePeriodSeconds: &s.Config.Kubernetes.TerminationGracePeriodSeconds,
 			ImagePullSecrets:              imagePullSecrets,
+			SecurityContext:               s.Config.Kubernetes.GetPodSecurityContext(),
 		},
-	})
+	}
+}
 
-	if err != nil {
-		return err
+func (s *executor) getHelperImage() string {
+	if len(s.Config.Kubernetes.HelperImage) > 0 {
+		return common.AppVersion.Variables().ExpandValue(s.Config.Kubernetes.HelperImage)
 	}
 
-	s.pod = pod
+	return s.helperImageInfo.String()
+}
 
-	return nil
+func (s *executor) makePodProxyServices() ([]api.Service, error) {
+	ch := make(chan serviceCreateResponse)
+	var wg sync.WaitGroup
+	wg.Add(len(s.ProxyPool))
+
+	for serviceName, serviceProxy := range s.ProxyPool {
+		serviceName := dns.MakeRFC1123Compatible(serviceName)
+		servicePorts := make([]api.ServicePort, len(serviceProxy.Settings.Ports))
+		for i, port := range serviceProxy.Settings.Ports {
+			// When there is more than one port Kubernetes requires a port name
+			portName := fmt.Sprintf("%s-%d", serviceName, port.Number)
+			servicePorts[i] = api.ServicePort{Port: int32(port.Number), TargetPort: intstr.FromInt(port.Number), Name: portName}
+		}
+
+		serviceConfig := s.prepareServiceConfig(serviceName, servicePorts)
+		go s.createKubernetesService(&serviceConfig, serviceProxy.Settings, ch, &wg)
+	}
+
+	go func() {
+		wg.Wait()
+		close(ch)
+	}()
+
+	var services []api.Service
+	for res := range ch {
+		if res.err != nil {
+			err := fmt.Errorf("error creating the proxy service %q: %w", res.service.Name, res.err)
+			s.Errorln(err)
+
+			return []api.Service{}, err
+		}
+
+		services = append(services, *res.service)
+	}
+
+	return services, nil
+}
+
+func (s *executor) prepareServiceConfig(name string, ports []api.ServicePort) api.Service {
+	return api.Service{
+		ObjectMeta: metav1.ObjectMeta{
+			GenerateName: name,
+			Namespace:    s.configurationOverwrites.namespace,
+		},
+		Spec: api.ServiceSpec{
+			Ports:    ports,
+			Selector: map[string]string{"pod": s.projectUniqueName()},
+			Type:     api.ServiceTypeClusterIP,
+		},
+	}
+}
+
+func (s *executor) createKubernetesService(service *api.Service, proxySettings *proxy.Settings, ch chan<- serviceCreateResponse, wg *sync.WaitGroup) {
+	defer wg.Done()
+
+	service, err := s.kubeClient.CoreV1().Services(s.pod.Namespace).Create(service)
+	if err == nil {
+		// Updating the internal service name reference and activating the proxy
+		proxySettings.ServiceName = service.Name
+	}
+
+	ch <- serviceCreateResponse{service: service, err: err}
 }
 
 func (s *executor) runInContainer(ctx context.Context, name string, command []string, script string) <-chan error {
@@ -498,10 +673,22 @@ func (s *executor) prepareOverwrites(variables common.JobVariables) error {
 	return nil
 }
 
-func (s *executor) prepareOptions(job *common.Build) {
+func (s *executor) prepareOptions(build *common.Build) {
 	s.options = &kubernetesOptions{}
-	s.options.Image = job.Image
-	for _, service := range job.Services {
+	s.options.Image = build.Image
+
+	s.getServices(build)
+}
+
+func (s *executor) getServices(build *common.Build) {
+	for _, service := range s.Config.Kubernetes.Services {
+		if service.Name == "" {
+			continue
+		}
+		s.options.Services = append(s.options.Services, common.Image{Name: service.Name})
+	}
+
+	for _, service := range build.Services {
 		if service.Name == "" {
 			continue
 		}
@@ -532,10 +719,19 @@ func (s *executor) checkDefaults() error {
 }
 
 func createFn() common.Executor {
+	helperImageInfo, err := helperimage.Get(common.REVISION, helperimage.Config{
+		OSType:       helperimage.OSTypeLinux,
+		Architecture: "amd64",
+	})
+	if err != nil {
+		logrus.WithError(err).Fatal("Failed to set up helper image for kubernetes executor")
+	}
+
 	return &executor{
 		AbstractExecutor: executors.AbstractExecutor{
 			ExecutorOptions: executorOptions,
 		},
+		helperImageInfo: helperImageInfo,
 	}
 }
 
@@ -545,6 +741,9 @@ func featuresFn(features *common.FeaturesInfo) {
 	features.Services = true
 	features.Artifacts = true
 	features.Cache = true
+	features.Session = true
+	features.Terminal = true
+	features.Proxy = true
 }
 
 func init() {
